==> rasa_x-0.42.0.dev20210713-py3-none-any.whl: rasax/community/services/settings_service.py <==
import json
import logging
from typing import Dict, Text, Any, Optional, Union, List

import typing
from ruamel.yaml.compat import ordereddict
from sanic.request import Request
from sqlalchemy import and_
from sqlalchemy.orm import Session

import rasax.community.config as rasa_x_config
import rasax.community.constants as constants
import rasax.community.utils.cli as cli_utils
import rasax.community.utils.config as config_utils
import rasax.community.utils.io as io_utils
import rasax.community.utils.yaml as yaml_utils
from rasax.community.database.admin import Environment, Project
from rasax.community.database.service import DbService
from rasax.community.services import background_dump_service
from rasax.community.services.data_service import DataService
from rasax.community.services.domain_service import DomainService
from rasax.community.services.story_service import StoryService
from rasax.community.services.license_service import LicenseService

if typing.TYPE_CHECKING:
    from rasax.community.services.stack_service import StackService

logger = logging.getLogger(__name__)


class RasaXEnvironmentError(ValueError):
    """Raised if the deployment environment doesn't match the requirements.

    An example for this is if no Rasa Open Source deployment is available.
    """


def default_environments_config_local(rasa_port: Union[int, Text]) -> Dict[Text, Any]:
    """Returns the default environment config for local mode."""
    stack_config = ordereddict(
        [
            (
                constants.RASA_PRODUCTION_ENVIRONMENT,
                ordereddict(
                    [
                        ("url", f"http://localhost:{rasa_port}"),
                        ("token", rasa_x_config.rasa_token),
                    ]
                ),
            ),
            (
                constants.RASA_DEVELOPMENT_ENVIRONMENT,
                ordereddict(
                    [
                        ("url", f"http://stuff:{rasa_port}"),
                        ("token", rasa_x_config.rasa_token),
                    ]
                ),
            ),
            (
                constants.RASA_WORKER_ENVIRONMENT,
                ordereddict(
                    [
                        ("url", f"http://localhost:{rasa_port}"),
                        ("token", rasa_x_config.rasa_token),
                    ]
                ),
            ),
        ]
    )

    return {"environments": ordereddict([("rasa", stack_config)])}


def default_stack_config() -> Dict[Text, Union[Text, List[Dict[Text, Text]]]]:
    # TODO: Find a way to keep this in sync with the defaults that Rasa OSS
    # chooses?

    pipeline = [
        {"name": "WhitespaceTokenizer"},
        {"name": "RegexFeaturizer"},
        {"name": "LexicalSyntacticFeaturizer"},
        {"name": "CountVectorsFeaturizer"},
        {
            "name": "CountVectorsFeaturizer",
            "analyzer": "char_wb",
            "min_ngram": 1,
            "max_ngram": 4,
        },
        {"name": "DIETClassifier", "epochs": 100},
        {"name": "EntitySynonymMapper"},
        {"name": "ResponseSelector", "epochs": 100},
    ]

    policies = [
        {"name": "MemoizationPolicy"},
        {"name": "TEDPolicy", "max_history": 5, "epochs": 100},
        {"name": "RulePolicy"},
    ]

    return ordereddict(
        [("language", "en"), ("pipeline", pipeline), ("policies", policies)]
    )


class ProjectException(Exception):
    """Exception raised for errors related to projects."""

    def __init__(self, username: Text):
        self.message = username

    def __str__(self) -> Text:
        return self.message


class SettingsService(DbService):
    """A service used to read/write configuration of the Rasa X project."""

    @staticmethod
    def from_request(request: Request, **kwargs) -> "SettingsService":
        """Creates a `SettingsService` from an incoming HTTP request's DB session.

        Args:
            request: Incoming HTTP request.
            **kwargs: other key-value args, not used.

        Returns:
            `SettingsService` instance with a connection to the DB.
        """
        return SettingsService(request.ctx.db_session)

    def init_project(self, team: Text, project_id: Text):
        """Create project for `project_id`.

        Raise `ProjectException` if a project under that name already exists.
        """

        if self.get(team, project_id):
            raise ProjectException(f"Project '{project_id}' already exists.")

        new_project = Project(
            project_id=project_id, team=team, config=json.dumps(default_stack_config())
        )
        self.add(new_project)

        return new_project.as_dict()

    def get(
        self, team: Text, project_id: Text
    ) -> Optional[Dict[Text, Union[Text, Dict]]]:
        project = self._get_project(team, project_id)
        return project.as_dict() if project else None

    def get_config(
        self, team: Text, project_id: Text
    ) -> Optional[Dict[Text, Union[Text, Dict]]]:
        """Get model config associated with project."""

        project = self._get_project(team, project_id)
        return project.get_model_config() if project else None

    def _get_project(self, team: Text, project_id: Text) -> Optional[Project]:
        return (
            self.query(Project)
            .filter(and_(Project.project_id == project_id, Project.team == team))
            .first()
        )

    def dump_config(
        self, team: Text, project_id: Text, filename: Optional[Text] = None
    ) -> None:
        """Dump domain to `filename` in yml format.

        Args:
            team: Team whose config should be dumped.
            project_id: Project whose config should be dumped.
            filename: Name of the file the config should be dumped into.
        """

        project = self._get_project(team, project_id)
        if project:
            if not filename:
                default_path = str(
                    io_utils.get_project_directory() / rasa_x_config.default_config_path
                )
                filename = json.loads(project.config).get("path") or default_path
            yaml_utils.dump_yaml_to_file(
                filename=filename, content=project.get_model_config()
            )

    def save_config(
        self,
        team: Text,
        project_id: Text,
        stack_config: Dict,
        config_path: Optional[Text] = None,
        should_dump: bool = True,
    ) -> None:
        logger.debug(stack_config)

        if config_path:
            stack_config["path"] = config_path

        project = self._get_project(team, project_id)
        project.config = json.dumps(stack_config)

        if should_dump:
            background_dump_service.add_model_configuration_change(team, project_id)

    @staticmethod
    def _inspect_config(stack_config: Dict) -> None:
        """Confirm Rasa config has all the mandatory keys."""

        missing_keys = [
            k
            for k in constants.RASA_CONFIG_MANDATORY_KEYS
            if k not in stack_config or stack_config[k] is None
        ]
        if missing_keys:
            raise config_utils.InvalidConfigError(
                "The config file is missing mandatory parameters: "
                "'{}'. Add missing parameters to config file and try again."
                "".format("', '".join(missing_keys))
            )

    def inspect_and_load_yaml_config(self, config_yaml: Text) -> Dict:
        stack_config = yaml_utils.load_yaml(config_yaml)
        self._inspect_config(stack_config)

        return stack_config

    def inspect_and_save_yaml_config_from_request(
        self, request_body: bytes, team: Text, project_id: Text
    ) -> Text:
        """Inspect and save yaml config from `request_body`."""

        config_yaml = io_utils.convert_bytes_to_string(request_body)
        stack_config = self.inspect_and_load_yaml_config(config_yaml)

        self.save_config(team, project_id, stack_config)

        return config_yaml

    def inject_environments_config_from_file(self, project_id: Text, filename: Text):
        """Inject a deployment environments configuration file at `filename`
        into the db.
        """

        try:
            _config = yaml_utils.read_yaml_file(filename)
            self.inspect_environments_config(_config, self.session)
            self.save_environments_config(project_id, _config)
            logger.debug(
                f"Successfully injected deployment environments "
                f"configuration from file '{filename}'."
            )
        except (ValueError, FileNotFoundError) as e:
            current_config = self.get_environments_config(project_id)
            logger.warning(
                f"Could not inject deployment environments "
                f"configuration from file '{filename}'. Details:\n{e}"
            )

            if current_config:
                logger.warning(
                    f"Rasa X will continue to use your current configuration."
                    f"Your current configuration is:\n{json.dumps(current_config, indent=2)}"
                )
            else:
                raise RasaXEnvironmentError(
                    "Could not get valid environments for server startup."
                )

    @staticmethod
    def inspect_environments_config(
        env_config: Dict[Text, Any], session: Optional[Session] = None
    ) -> None:
        """Inspect deployment environments config.

        Check whether required entries `worker` and `production` are available, and if any
        extra entries added should be allowed. Remove unallowed entries.

        Args:
            env_config: The environments config to inspect.
            session: Optional SQLAlchemy session to use when checking for Enterprise
                license.

        Raises:
            RasaXEnvironmentError: Raised if config is invalid because it does not contain required entries
                for worker or production services.
        """
        rasa_config = env_config.get("rasa")

        if not rasa_config:
            raise RasaXEnvironmentError("Environment config needs to have key `rasa`.")

        if any(key not in rasa_config for key in constants.ENVIRONMENTS_MANDATORY_KEYS):
            raise RasaXEnvironmentError(
                "Environment needs to contain entries for production "
                "and worker services."
            )

        if LicenseService.is_enterprise_activated(session):
            return

        community_unallowed_keys = [
            key
            for key in rasa_config.keys()
            if key not in constants.ENVIRONMENTS_MANDATORY_KEYS
        ]
        if community_unallowed_keys:
            cli_utils.raise_warning(
                f"Rasa X only allows for a production environment and "
                f"a worker environment. Rasa X will start up with only these "
                f"environments. To use more environments, please "
                f"contact us at {constants.HI_RASA_EMAIL} for a "
                f"Rasa Enterprise license."
            )
            for key in community_unallowed_keys:
                del rasa_config[key]

    def save_environments_config(
        self, project_id: Text, environments_config: Dict[Text, Any]
    ) -> Dict[Text, Dict[Text, Dict[Text, Text]]]:
        """Save the provided environments configuration to the project, replacing old environments.

        Args:
            project_id: The project to apply the environments to.
            environments_config: The environments configuration to save to the project

        Returns:
            The environments config stored under the `project_id` after saving.
        """
        self.query(Environment).delete()

        for name, env_config in environments_config.get("rasa", {}).items():
            env = Environment(
                name=name,
                project=project_id,
                url=env_config["url"],
                token=env_config["token"],
            )
            self.add(env)

        return self.get_environments_config(project_id)

    def get_environments_config(self, project_id: Text) -> Dict[Text, Any]:
        """Retrieve the environments configuration from the database.

        Args:
            project_id: The project to retrieve environments from.

        Returns:
            The environments config stored for the given `project_id`.
        """
        envs = self.query(Environment).filter(Environment.project == project_id).all()
        envs = [e.as_dict() for e in envs]

        return {"environments": {"rasa": {list(e)[0]: e[list(e)[0]] for e in envs}}}

    def get_stack_service(
        self, environment: Text, project_id: Text = rasa_x_config.project_name
    ) -> Optional["StackService"]:
        return self.stack_services(project_id).get(environment, None)

    def stack_services(
        self, project_id: Text = rasa_x_config.project_name
    ) -> Dict[Text, "StackService"]:
        """Create StackServices for all Stack servers."""
        from rasax.community.services.stack_service import RasaCredentials
        from rasax.community.services.stack_service import StackService

        environments_config = self.get_environments_config(project_id).get(
            "environments", {}
        )

        _stack_services = {}
        for k, v in environments_config.get("rasa", {}).items():
            credentials = RasaCredentials(url=v["url"], token=v.get("token"))
            _stack_services[k] = StackService(
                credentials,
                DataService(self.session),
                StoryService(self.session),
                DomainService(self.session),
                self,
            )

        return _stack_services

==> rasa_x-0.42.3-py3-none-any.whl: rasax/community/services/settings_service.py <==
import json
import logging
from typing import Dict, Text, Any, Optional, Union, List

import typing
from ruamel.yaml.compat import ordereddict
from sanic.request import Request
from sqlalchemy import and_

import rasax.community.config as rasa_x_config
import rasax.community.constants as constants
import rasax.community.utils.cli as cli_utils
import rasax.community.utils.config as config_utils
import rasax.community.utils.common as common_utils
import rasax.community.utils.io as io_utils
import rasax.community.utils.yaml as yaml_utils
from rasax.community.database.admin import Environment, Project, LocalPassword
from rasax.community.database.service import DbService
from rasax.community.services import background_dump_service
from rasax.community.services.data_service import DataService
from rasax.community.services.domain_service import DomainService
from rasax.community.services.story_service import StoryService

if typing.TYPE_CHECKING:
    from rasax.community.services.stack_service import StackService

logger = logging.getLogger(__name__)


class RasaXEnvironmentError(ValueError):
    """Raised if the deployment environment doesn't match the requirements.

    An example for this is if no Rasa Open Source deployment is available.
    """


def default_environments_config_local(rasa_port: Union[int, Text]) -> Dict[Text, Any]:
    """Returns the default environment config for local mode."""
    stack_config = ordereddict(
        [
            (
                constants.RASA_PRODUCTION_ENVIRONMENT,
                ordereddict(
                    [
                        ("url", f"http://localhost:{rasa_port}"),
                        ("token", rasa_x_config.rasa_token),
                    ]
                ),
            ),
            (
                constants.RASA_DEVELOPMENT_ENVIRONMENT,
                ordereddict(
                    [
                        ("url", f"http://stuff:{rasa_port}"),
                        ("token", rasa_x_config.rasa_token),
                    ]
                ),
            ),
            (
                constants.RASA_WORKER_ENVIRONMENT,
                ordereddict(
                    [
                        ("url", f"http://localhost:{rasa_port}"),
                        ("token", rasa_x_config.rasa_token),
                    ]
                ),
            ),
        ]
    )

    return {"environments": ordereddict([("rasa", stack_config)])}


def default_stack_config() -> Dict[Text, Union[Text, List[Dict[Text, Text]]]]:
    # TODO: Find a way to keep this in sync with the defaults that Rasa OSS
    # chooses?

    pipeline = [
        {"name": "WhitespaceTokenizer"},
        {"name": "RegexFeaturizer"},
        {"name": "LexicalSyntacticFeaturizer"},
        {"name": "CountVectorsFeaturizer"},
        {
            "name": "CountVectorsFeaturizer",
            "analyzer": "char_wb",
            "min_ngram": 1,
            "max_ngram": 4,
        },
        {"name": "DIETClassifier", "epochs": 100},
        {"name": "EntitySynonymMapper"},
        {"name": "ResponseSelector", "epochs": 100},
    ]

    policies = [
        {"name": "MemoizationPolicy"},
        {"name": "TEDPolicy", "max_history": 5, "epochs": 100},
        {"name": "RulePolicy"},
    ]

    return ordereddict(
        [("language", "en"), ("pipeline", pipeline), ("policies", policies)]
    )


class ProjectException(Exception):
    """Exception raised for errors related to projects."""

    def __init__(self, username: Text):
        self.message = username

    def __str__(self) -> Text:
        return self.message


class SettingsService(DbService):
    """A service used to read/write configuration of the Rasa X project."""

    @staticmethod
    def from_request(request: Request, **kwargs) -> "SettingsService":
        """Creates a `SettingsService` from an incoming HTTP request's DB session.

        Args:
            request: Incoming HTTP request.
            **kwargs: other key-value args, not used.

        Returns:
            `SettingsService` instance with a connection to the DB.
        """
        return SettingsService(request.ctx.db_session)

    def init_project(self, team: Text, project_id: Text):
        """Create project for `project_id`.

        Raise `ProjectException` if a project under that name already exists.
        """

        if self.get(team, project_id):
            raise ProjectException(f"Project '{project_id}' already exists.")

        new_project = Project(
            project_id=project_id, team=team, config=json.dumps(default_stack_config())
        )
        self.add(new_project)

        return new_project.as_dict()

    def get(
        self, team: Text, project_id: Text
    ) -> Optional[Dict[Text, Union[Text, Dict]]]:
        project = self._get_project(team, project_id)
        return project.as_dict() if project else None

    def get_config(
        self, team: Text, project_id: Text
    ) -> Optional[Dict[Text, Union[Text, Dict]]]:
        """Get model config associated with project."""

        project = self._get_project(team, project_id)
        return project.get_model_config() if project else None

    def _get_project(self, team: Text, project_id: Text) -> Optional[Project]:
        return (
            self.query(Project)
            .filter(and_(Project.project_id == project_id, Project.team == team))
            .first()
        )

    def dump_config(
        self, team: Text, project_id: Text, filename: Optional[Text] = None
    ) -> None:
        """Dump domain to `filename` in yml format.

        Args:
            team: Team whose config should be dumped.
            project_id: Project whose config should be dumped.
            filename: Name of the file the config should be dumped into.
        """

        project = self._get_project(team, project_id)
        if project:
            if not filename:
                default_path = str(
                    io_utils.get_project_directory() / rasa_x_config.default_config_path
                )
                filename = json.loads(project.config).get("path") or default_path
            yaml_utils.dump_yaml_to_file(
                filename=filename, content=project.get_model_config()
            )

    def save_config(
        self,
        team: Text,
        project_id: Text,
        stack_config: Dict,
        config_path: Optional[Text] = None,
        should_dump: bool = True,
    ) -> None:
        logger.debug(stack_config)

        if config_path:
            stack_config["path"] = config_path

        project = self._get_project(team, project_id)
        project.config = json.dumps(stack_config)

        if should_dump:
            background_dump_service.add_model_configuration_change(team, project_id)

    @staticmethod
    def _inspect_config(stack_config: Dict) -> None:
        """Confirm Rasa config has all the mandatory keys."""

        missing_keys = [
            k
            for k in constants.RASA_CONFIG_MANDATORY_KEYS
            if k not in stack_config or stack_config[k] is None
        ]
        if missing_keys:
            raise config_utils.InvalidConfigError(
                "The config file is missing mandatory parameters: "
                "'{}'. Add missing parameters to config file and try again."
                "".format("', '".join(missing_keys))
            )

    def inspect_and_load_yaml_config(self, config_yaml: Text) -> Dict:
        stack_config = yaml_utils.load_yaml(config_yaml)
        self._inspect_config(stack_config)

        return stack_config

    def inspect_and_save_yaml_config_from_request(
        self, request_body: bytes, team: Text, project_id: Text
    ) -> Text:
        """Inspect and save yaml config from `request_body`."""

        config_yaml = io_utils.convert_bytes_to_string(request_body)
        stack_config = self.inspect_and_load_yaml_config(config_yaml)

        self.save_config(team, project_id, stack_config)

        return config_yaml

    def inject_environments_config_from_file(self, project_id: Text, filename: Text):
        """Inject a deployment environments configuration file at `filename`
        into the db.
        """

        try:
            _config = yaml_utils.read_yaml_file(filename)
            self.inspect_environments_config(_config)
            self.save_environments_config(project_id, _config)
            logger.debug(
                f"Successfully injected deployment environments "
                f"configuration from file '{filename}'."
            )
        except (ValueError, FileNotFoundError) as e:
            current_config = self.get_environments_config(project_id)
            logger.warning(
                f"Could not inject deployment environments "
                f"configuration from file '{filename}'. Details:\n{e}"
            )

            if current_config:
                logger.warning(
                    f"Rasa X will continue to use your current configuration."
                    f"Your current configuration is:\n{json.dumps(current_config, indent=2)}"
                )
            else:
                raise RasaXEnvironmentError(
                    "Could not get valid environments for server startup."
                )

    @staticmethod
    def inspect_environments_config(env_config: Dict[Text, Any]) -> None:
        """Inspect deployment environments config.

        Check whether required entries `worker` and `production` are available, and if any
        extra entries added should be allowed. Remove unallowed entries.

        Args:
            env_config: The environments config to inspect.

        Raises:
            RasaXEnvironmentError: Raised if config is invalid because it does not contain required entries
                for worker or production services.
        """
        rasa_config = env_config.get("rasa")

        if not rasa_config:
            raise RasaXEnvironmentError("Environment config needs to have key `rasa`.")

        if any(key not in rasa_config for key in constants.ENVIRONMENTS_MANDATORY_KEYS):
            raise RasaXEnvironmentError(
                "Environment needs to contain entries for production "
                "and worker services."
            )

        if common_utils.is_enterprise_installed():
            return

        community_unallowed_keys = [
            key
            for key in rasa_config.keys()
            if key not in constants.ENVIRONMENTS_MANDATORY_KEYS
        ]
        if community_unallowed_keys:
            cli_utils.raise_warning(
                f"Rasa X only allows for a production environment and "
                f"a worker environment. Rasa X will start up with only these "
                f"environments. To use more environments, please "
                f"contact us at {constants.HI_RASA_EMAIL} for a "
                f"Rasa Enterprise license."
            )
            for key in community_unallowed_keys:
                del rasa_config[key]

    def save_environments_config(
        self, project_id: Text, environments_config: Dict[Text, Any]
    ) -> Dict[Text, Dict[Text, Dict[Text, Text]]]:
        """Save the provided environments configuration to the project.

        Args:
            project_id: The project to apply the enviornments to.
            environments_config: The environments configuration to save to the project

        Returns:
            The environments config stored under the `project_id` after saving.
        """
        for name, env_config in environments_config.get("rasa", {}).items():
            # delete possibly existing config
            self.delete_environment_config(name, project_id)

            env = Environment(
                name=name,
                project=project_id,
                url=env_config["url"],
                token=env_config["token"],
            )
            self.add(env)

        return self.get_environments_config(project_id)

    def delete_environment_config(self, name: Text, project_id: Text) -> None:
        to_delete = (
            self.query(Environment)
            .filter(Environment.name == name)
            .filter(Environment.project == project_id)
            .first()
        )
        if to_delete:
            self.delete(to_delete)

    def get_environments_config(self, project_id: Text) -> Dict[Text, Any]:
        envs = self.query(Environment).filter(Environment.project == project_id).all()
        envs = [e.as_dict() for e in envs]

        return {"environments": {"rasa": {list(e)[0]: e[list(e)[0]] for e in envs}}}

    def get_stack_service(
        self, environment: Text, project_id: Text = rasa_x_config.project_name
    ) -> Optional["StackService"]:
        return self.stack_services(project_id).get(environment, None)

    def stack_services(
        self, project_id: Text = rasa_x_config.project_name
    ) -> Dict[Text, "StackService"]:
        """Create StackServices for all Stack servers."""
        from rasax.community.services.stack_service import RasaCredentials
        from rasax.community.services.stack_service import StackService

        environments_config = self.get_environments_config(project_id).get(
            "environments", {}
        )

        _stack_services = {}
        for k, v in environments_config.get("rasa", {}).items():
            credentials = RasaCredentials(url=v["url"], token=v.get("token"))
            _stack_services[k] = StackService(
                credentials,
                DataService(self.session),
                StoryService(self.session),
                DomainService(self.session),
                self,
            )

        return _stack_services

    def save_community_user_password(self, password: Text) -> None:
        """Save Rasa X password in local mode.

        Overwrite existing local password, or create a new one if none is found.
        """

        local_password = self.query(LocalPassword).first()
        if local_password:
            local_password.password = password
        else:
            local_password = LocalPassword(password=password)
            self.add(local_password)

    def get_community_user_password(self) -> Optional[Text]:
        """Fetch Rasa X password in local mode."""

        local_password = self.query(LocalPassword).first()
        if local_password:
            return local_password.password

        return None
